#include "imu_9dof_dma.h"
#include "imu_int.h"
#include "ADXL345_accel.h"
#include "ITG3200_gyro.h"
#include "HMC5843_compass.h"
#include "ram.h"

// I2C private variables
volatile uint8_t i2c_error;
volatile i2c_op_t i2c_op;
uint8_t i2c_rd_addr;
uint8_t i2c_wr_addr;
uint8_t i2c_data_out[32];
uint8_t imu_stop_flag;
volatile uint8_t imu_stopped;
// dma initialization structures;
DMA_InitTypeDef IMU_DMA_TX_InitStructure;
DMA_InitTypeDef IMU_DMA_RX_InitStructure;

// private functions
void imu_wait_op_done(void)
{
  while(i2c_op!=I2C_OP_DONE);
}

uint8_t imu_is_op_done(void)
{
  if(i2c_op==I2C_OP_DONE)
    return 0x01;
  else
    return 0x00;
}

// interrupt handlers
void IMU_TIMER_IRQHandler(void)
{
  static imu_state_t state=IMU_GET_COMPASS;
  uint16_t capture;

  if(TIM_GetITStatus(IMU_TIMER, TIM_IT_CC1)!=RESET)
  {
    TIM_ClearITPendingBit(IMU_TIMER,TIM_IT_CC1);
    capture = TIM_GetCapture1(IMU_TIMER);
    TIM_SetCompare1(IMU_TIMER, capture + 1000);
    if(imu_is_op_done())
    {
      if(imu_stop_flag==0x01)
      {
        // stop the timer to get the imu data
        TIM_Cmd(IMU_TIMER, DISABLE);
        TIM_ITConfig(IMU_TIMER, TIM_IT_CC1, DISABLE);
  
        ram_clear_bit(BIOLOID_IMU_STATUS,0);
        imu_stop_flag=0x00;
        imu_stopped=0x01;
      }
      else
      {
        switch(state)
        {
          case IMU_GET_ACCEL: imu_accel_get_data();
                              imu_compass_process_data();
                              state=IMU_GET_GYRO;
                              break;
          case IMU_GET_GYRO: imu_gyro_get_data();
                             imu_accel_process_data();
                             state=IMU_GET_COMPASS;
                             break;
          case IMU_GET_COMPASS: imu_compass_get_data();
                                imu_gyro_process_data();
                                state=IMU_GET_ACCEL;
                                break;
        }
      }
    }
  }
}

void I2C_IMU_EV_IRQHandler(void)
{
  switch (I2C_GetLastEvent(I2C_IMU))
  {
    /* Test on I2C1 EV5 and clear it */
    case I2C_EVENT_MASTER_MODE_SELECT:              
      if(i2c_op==I2C_RD_OP_DATA)
        I2C_Send7bitAddress(I2C_IMU,i2c_rd_addr,I2C_Direction_Receiver);
      else
        I2C_Send7bitAddress(I2C_IMU,i2c_wr_addr,I2C_Direction_Transmitter);
      break;

    /* Test on I2C1 EV6 and first EV8 and clear them */
    case I2C_EVENT_MASTER_TRANSMITTER_MODE_SELECTED:
      I2C_ITConfig(I2C_IMU, I2C_IT_EVT, DISABLE);
      /* enable DMA transfer */
      I2C_DMACmd(I2C_IMU, ENABLE);
      DMA_Cmd(I2C_IMU_DMA_STREAM_TX, ENABLE);
      break;

    case I2C_EVENT_MASTER_RECEIVER_MODE_SELECTED:
      I2C_ITConfig(I2C_IMU, I2C_IT_EVT, DISABLE);
      /* enable DMA transfer */
      if(IMU_DMA_RX_InitStructure.DMA_BufferSize==1)
        I2C_AcknowledgeConfig(I2C_IMU,DISABLE);
      else
        I2C_AcknowledgeConfig(I2C_IMU,ENABLE);
      I2C_DMALastTransferCmd(I2C_IMU,ENABLE);
      DMA_Cmd(I2C_IMU_DMA_STREAM_RX, ENABLE);
      break;

    case I2C_EVENT_MASTER_BYTE_TRANSMITTED:
      if(i2c_op==I2C_WR_OP)
      {
        i2c_op=I2C_OP_DONE;
        I2C_GenerateSTOP(I2C_IMU, ENABLE);
      }
      break;

    default:
      break;
  }
}

void I2C_IMU_ER_IRQHandler(void)
{
  if(I2C_GetFlagStatus(I2C_IMU,I2C_FLAG_AF)!=RESET)
  {
    i2c_error=0x01;
    I2C_ClearFlag(I2C_IMU,I2C_FLAG_AF);
    if(i2c_op==I2C_RD_OP_DATA)
      DMA_Cmd(I2C_IMU_DMA_STREAM_RX, DISABLE);
    else
      DMA_Cmd(I2C_IMU_DMA_STREAM_TX, DISABLE);
    I2C_DMACmd(I2C_IMU, DISABLE);
    I2C_GenerateSTOP(I2C_IMU, ENABLE);
    i2c_op=I2C_OP_DONE;
  }
  I2C_ClearITPendingBit(I2C_IMU,I2C_IT_ERR);
}

void I2C_IMU_TX_DMA_IRQHandler(void)
{
  DMA_ClearFlag(I2C_IMU_DMA_STREAM_TX,I2C_IMU_TX_DMA_TCFLAG);
  DMA_ClearITPendingBit(I2C_IMU_DMA_STREAM_TX,I2C_IMU_TX_DMA_TCFLAG);
  I2C_ITConfig(I2C_IMU, I2C_IT_EVT, ENABLE);
  if(i2c_op==I2C_RD_OP_CMD)
  {
    I2C_GenerateSTART(I2C_IMU, ENABLE);
    i2c_op=I2C_RD_OP_DATA;
  }
}

void I2C_IMU_RX_DMA_IRQHandler(void)
{
  DMA_ClearFlag(I2C_IMU_DMA_STREAM_RX,I2C_IMU_RX_DMA_TCFLAG);
  DMA_ClearITPendingBit(I2C_IMU_DMA_STREAM_RX,I2C_IMU_RX_DMA_TCFLAG);
  I2C_ITConfig(I2C_IMU, I2C_IT_EVT, ENABLE);
  I2C_GenerateSTOP(I2C_IMU, ENABLE);
  i2c_op=I2C_OP_DONE;
}

// public functions
void imu_init(void)
{
  uint8_t i;
  TIM_TimeBaseInitTypeDef  TIM_TimeBaseStructure;
  TIM_OCInitTypeDef  TIM_OCInitStructure;
  GPIO_InitTypeDef  GPIO_InitStructure;
  NVIC_InitTypeDef NVIC_InitStructure;
  I2C_InitTypeDef  I2C_InitStructure;

  /* RCC Configuration */
  /*I2C Peripheral clock enable */
  RCC_APB1PeriphClockCmd(I2C_IMU_CLK, ENABLE);
  /*SDA GPIO clock enable */
  RCC_AHB1PeriphClockCmd(I2C_IMU_SDA_GPIO_CLK, ENABLE);
  /*SCL GPIO clock enable */
  RCC_AHB1PeriphClockCmd(I2C_IMU_SCL_GPIO_CLK, ENABLE);
  /* Enable the DMA clock */
  RCC_AHB1PeriphClockCmd(DMA_IMU_CLK, ENABLE);
  /* Enable the IMU timer clock */
  RCC_APB1PeriphClockCmd(IMU_TIMER_CLK,ENABLE);

  /* GPIO Configuration */
  /*Configure I2C SCL pin */
  GPIO_InitStructure.GPIO_Pin = I2C_IMU_SCL_PIN;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_OD;
  GPIO_InitStructure.GPIO_PuPd  = GPIO_PuPd_UP;
  GPIO_Init(I2C_IMU_SCL_GPIO_PORT, &GPIO_InitStructure);

  /*Configure I2C SDA pin */
  GPIO_InitStructure.GPIO_Pin = I2C_IMU_SDA_PIN;
  GPIO_InitStructure.GPIO_PuPd  = GPIO_PuPd_UP;
  GPIO_Init(I2C_IMU_SDA_GPIO_PORT, &GPIO_InitStructure);

  /* Connect PXx to I2C_SCL */
  GPIO_PinAFConfig(I2C_IMU_SCL_GPIO_PORT, I2C_IMU_SCL_SOURCE, I2C_IMU_SCL_AF);
  /* Connect PXx to I2C_SDA */
  GPIO_PinAFConfig(I2C_IMU_SDA_GPIO_PORT, I2C_IMU_SDA_SOURCE, I2C_IMU_SDA_AF);

  /* I2C Struct Initialize */
  I2C_InitStructure.I2C_Mode = I2C_Mode_I2C;
  I2C_InitStructure.I2C_DutyCycle = I2C_DutyCycle_2;
  I2C_InitStructure.I2C_OwnAddress1 = 0x00;
  I2C_InitStructure.I2C_Ack = I2C_Ack_Enable;
  I2C_InitStructure.I2C_ClockSpeed = 400000;
  I2C_InitStructure.I2C_AcknowledgedAddress = I2C_AcknowledgedAddress_7bit;
  I2C_Init(I2C_IMU, &I2C_InitStructure);
  /* configure I2C interrupts */
  NVIC_InitStructure.NVIC_IRQChannel = I2C_IMU_ER_IRQn;
  NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 3;
  NVIC_InitStructure.NVIC_IRQChannelSubPriority = 3;
  NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
  NVIC_Init(&NVIC_InitStructure);
  NVIC_InitStructure.NVIC_IRQChannel = I2C_IMU_EV_IRQn;
  NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 3;
  NVIC_InitStructure.NVIC_IRQChannelSubPriority = 3;
  NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
  NVIC_Init(&NVIC_InitStructure);
  I2C_ITConfig(I2C_IMU, I2C_IT_ERR | I2C_IT_EVT, ENABLE);
  // enable I2C1
  I2C_Cmd(I2C_IMU, ENABLE);

  // initialize data
  i2c_error=0x00;
  i2c_op=I2C_OP_DONE;
  imu_stop_flag=0x00;
  imu_stopped=0x00;
  for(i=0;i<32;i++)
    i2c_data_out[i]=0x00;
  /* configure DMA TX */
  DMA_DeInit(I2C_IMU_DMA_STREAM_TX);
  IMU_DMA_TX_InitStructure.DMA_FIFOMode = DMA_FIFOMode_Disable;
  IMU_DMA_TX_InitStructure.DMA_FIFOThreshold = DMA_FIFOThreshold_Full;
  IMU_DMA_TX_InitStructure.DMA_MemoryBurst = DMA_MemoryBurst_Single;
  IMU_DMA_TX_InitStructure.DMA_MemoryDataSize = DMA_MemoryDataSize_Byte;
  IMU_DMA_TX_InitStructure.DMA_MemoryInc = DMA_MemoryInc_Enable;
  IMU_DMA_TX_InitStructure.DMA_Mode = DMA_Mode_Normal;
  IMU_DMA_TX_InitStructure.DMA_PeripheralBaseAddr = I2C_IMU_DR_ADDRESS;
  IMU_DMA_TX_InitStructure.DMA_PeripheralBurst = DMA_PeripheralBurst_Single;
  IMU_DMA_TX_InitStructure.DMA_PeripheralDataSize = DMA_PeripheralDataSize_Byte;
  IMU_DMA_TX_InitStructure.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
  IMU_DMA_TX_InitStructure.DMA_Priority = DMA_Priority_High;
  IMU_DMA_TX_InitStructure.DMA_Channel = I2C_IMU_DMA_CHANNEL;
  IMU_DMA_TX_InitStructure.DMA_DIR = DMA_DIR_MemoryToPeripheral;
  /* initialize DMA interrupts */
  NVIC_InitStructure.NVIC_IRQChannel = I2C_IMU_TX_DMA_IRQn;
  NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 3;
  NVIC_InitStructure.NVIC_IRQChannelSubPriority = 1;
  NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
  NVIC_Init(&NVIC_InitStructure);
  DMA_ITConfig(I2C_IMU_DMA_STREAM_TX,DMA_IT_TC,ENABLE);
  DMA_ITConfig(I2C_IMU_DMA_STREAM_TX,DMA_IT_HT | DMA_IT_TE | DMA_IT_FE | DMA_IT_DME,DISABLE);

  /* configure DMA Rx */
  DMA_DeInit(I2C_IMU_DMA_STREAM_RX);
  IMU_DMA_RX_InitStructure.DMA_FIFOMode = DMA_FIFOMode_Disable;
  IMU_DMA_RX_InitStructure.DMA_FIFOThreshold = DMA_FIFOThreshold_Full;
  IMU_DMA_RX_InitStructure.DMA_MemoryBurst = DMA_MemoryBurst_Single;
  IMU_DMA_RX_InitStructure.DMA_MemoryDataSize = DMA_MemoryDataSize_Byte;
  IMU_DMA_RX_InitStructure.DMA_MemoryInc = DMA_MemoryInc_Enable;
  IMU_DMA_RX_InitStructure.DMA_Mode = DMA_Mode_Normal;
  IMU_DMA_RX_InitStructure.DMA_PeripheralBaseAddr = I2C_IMU_DR_ADDRESS;
  IMU_DMA_RX_InitStructure.DMA_PeripheralBurst = DMA_PeripheralBurst_Single;
  IMU_DMA_RX_InitStructure.DMA_PeripheralDataSize = DMA_PeripheralDataSize_Byte;
  IMU_DMA_RX_InitStructure.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
  IMU_DMA_RX_InitStructure.DMA_Priority = DMA_Priority_High;
  IMU_DMA_RX_InitStructure.DMA_Channel = I2C_IMU_DMA_CHANNEL;
  IMU_DMA_RX_InitStructure.DMA_DIR = DMA_DIR_PeripheralToMemory;
  /* initialize DMA interrupts */
  NVIC_InitStructure.NVIC_IRQChannel = I2C_IMU_RX_DMA_IRQn;
  NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 3;
  NVIC_InitStructure.NVIC_IRQChannelSubPriority = 2;
  NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
  NVIC_Init(&NVIC_InitStructure);
  DMA_ITConfig(I2C_IMU_DMA_STREAM_RX,DMA_IT_TC,ENABLE);
  DMA_ITConfig(I2C_IMU_DMA_STREAM_RX,DMA_IT_HT | DMA_IT_TE | DMA_IT_FE | DMA_IT_DME,DISABLE);

  // configure the timer
  // initialize the timer interrupts
  NVIC_InitStructure.NVIC_IRQChannel = IMU_TIMER_IRQn;
  NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 3;
  NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
  NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
  NVIC_Init(&NVIC_InitStructure);

  /* motion timer configuration */
  TIM_TimeBaseStructure.TIM_Period = 0xFFFF;
  TIM_TimeBaseStructure.TIM_Prescaler = 82;
  TIM_TimeBaseStructure.TIM_ClockDivision = 0;
  TIM_TimeBaseStructure.TIM_CounterMode = TIM_CounterMode_Up;
  TIM_TimeBaseInit(IMU_TIMER,&TIM_TimeBaseStructure);
  TIM_SetClockDivision(IMU_TIMER,TIM_CKD_DIV1);

  TIM_OCInitStructure.TIM_OCMode = TIM_OCMode_Timing;
  TIM_OCInitStructure.TIM_OutputState = TIM_OutputState_Disable;
  TIM_OCInitStructure.TIM_OCPolarity = TIM_OCPolarity_High;
  TIM_OCInitStructure.TIM_Pulse = 1000;// 1ms period
  TIM_OC1Init(IMU_TIMER, &TIM_OCInitStructure);
  TIM_OC1PreloadConfig(IMU_TIMER, TIM_OCPreload_Disable);

  /* detect the three sensors in the imu */
  if(imu_accel_detect())
    ram_set_bit(BIOLOID_IMU_STATUS,1);
  else
    ram_clear_bit(BIOLOID_IMU_STATUS,1);
  if(imu_gyro_detect())
    ram_set_bit(BIOLOID_IMU_STATUS,2);
  else
    ram_clear_bit(BIOLOID_IMU_STATUS,2);
  if(imu_compass_detect())
    ram_set_bit(BIOLOID_IMU_STATUS,3);
  else
    ram_clear_bit(BIOLOID_IMU_STATUS,3);
}

void imu_start(void)
{
  // start the accelerometer
  imu_accel_start();
  imu_gyro_start();
  imu_compass_start();

  // start the timer to get the imu data
  TIM_Cmd(IMU_TIMER, ENABLE);
  TIM_ITConfig(IMU_TIMER, TIM_IT_CC1, ENABLE);

  ram_set_bit(BIOLOID_IMU_STATUS,0);
}

void imu_stop(void)
{
  imu_stop_flag=0x01;
  while(!imu_stopped);
  imu_stopped=0x00;
  imu_accel_stop();
  imu_gyro_stop();
  imu_compass_stop();
}
