#include "motion_manager.h"
#include "dynamixel_master_uart_dma.h"
#include "dyn_servos.h"
#include "gpio.h"
#include "ram.h"
#include "action.h"
#include "stm32_time.h"

#define      MOTION_TIMER                 TIM3
#define      MOTION_TIMER_IRQn            TIM3_IRQn
#define      MOTION_TIMER_IRQHandler      TIM3_IRQHandler
#define      MOTION_TIMER_CLK             RCC_APB1Periph_TIM3

// private variables
__IO uint16_t manager_motion_period;
uint8_t manager_num_servos;
TServoInfo manager_servos[MANAGER_MAX_NUM_SERVOS];

// private functions
void manager_send_motion_command(void)
{
  uint8_t servo_ids[MANAGER_MAX_NUM_SERVOS];
  uint8_t data[256];
  uint8_t i,num=0;
  
  for(i=0;i<MANAGER_MAX_NUM_SERVOS;i++)
  {
    if(manager_servos[i].enabled)
    {
      servo_ids[num]=manager_servos[i].id;
      data[num*4]=action_slope[i];
      data[num*4+1]=action_slope[i];
      data[num*4+2]=manager_servos[i].current_value%256;
      data[num*4+3]=manager_servos[i].current_value/256;
      num++;
    }
  }
  if(num>0)
    dyn_master_sync_write(num,servo_ids,P_CW_COMPLIANCE_SLOPE,4,data);
}

inline uint16_t manager_angle_to_value(uint8_t servo_id,int16_t angle)
{
  return ((angle+manager_servos[servo_id].center_angle)*manager_servos[servo_id].encoder_resolution)/manager_servos[servo_id].max_angle;
}

inline int16_t manager_value_to_angle(uint8_t servo_id,uint16_t value)
{
  return (((int16_t)((value*manager_servos[servo_id].max_angle)/manager_servos[servo_id].encoder_resolution))-manager_servos[servo_id].center_angle);
}

inline uint16_t manager_speed_to_value(uint8_t servo_id,uint16_t speed)
{
  if(speed>manager_servos[servo_id].max_speed)
    speed=manager_servos[servo_id].max_speed;
  return (speed*3)>>1;
}

inline uint16_t manager_value_to_speed(uint8_t servo_id,uint16_t value)
{
  return (value*2)/3;
}

void manager_get_current_position(void)
{
  uint8_t i;

  for(i=0;i<MANAGER_MAX_NUM_SERVOS;i++)
  {
    if(!manager_servos[i].enabled)// servo is enabled
    {
      dyn_master_read_word(manager_servos[i].id,P_PRESENT_POSITION_L,&manager_servos[i].current_value);
      manager_servos[i].current_angle=manager_value_to_angle(i,manager_servos[i].current_value);
    }
    ram_data[BIOLOID_MM_SERVO0_CUR_POS_L+i*2]=(manager_servos[i].current_angle&0xFF);
    ram_data[BIOLOID_MM_SERVO0_CUR_POS_H+i*2]=(manager_servos[i].current_angle>>8);
  }
}

void manager_get_target_position(void)
{
  uint8_t i;

  for(i=0;i<MANAGER_MAX_NUM_SERVOS;i++)
  {
    if(manager_servos[i].enabled)// servo is enabled
    {
      if(manager_servos[i].module==MM_ACTION)
      {
        manager_servos[i].current_angle=((action_current_angles[i])>>9);
        //>>16 from the action codification, <<7 from the manager codification
        manager_servos[i].current_value=manager_angle_to_value(i,manager_servos[i].current_angle);
      }
    }
  }
}

void manager_balance(void)
{
  int16_t x_error=0.0;
  int16_t y_error=0.0;

  manager_servos[10].current_value+=(y_error*ram_data[BIOLOID_MM_PELVIS_ROLL_GAIN_L]*11)/2500;
  manager_servos[9].current_value+=(y_error*ram_data[BIOLOID_MM_PELVIS_ROLL_GAIN_R]*11)/2500;
  manager_servos[18].current_value-=(y_error*ram_data[BIOLOID_MM_ANKLE_ROLL_GAIN_L]*11)/2500;
  manager_servos[17].current_value-=(y_error*ram_data[BIOLOID_MM_ANKLE_ROLL_GAIN_R]*11)/2500;
  manager_servos[16].current_value-=(x_error*ram_data[BIOLOID_MM_ANKLE_PITCH_GAIN_L]*11)/2500;
  manager_servos[15].current_value+=(x_error*ram_data[BIOLOID_MM_ANKLE_PITCH_GAIN_R]*11)/2500;
  manager_servos[14].current_value-=(x_error*ram_data[BIOLOID_MM_KNEE_GAIN_L]*11)/2500;
  manager_servos[13].current_value+=(x_error*ram_data[BIOLOID_MM_KNEE_GAIN_R]*11)/2500;
}

// interrupt handlers
void MOTION_TIMER_IRQHandler(void)
{
  uint16_t capture;

  if(TIM_GetITStatus(MOTION_TIMER, TIM_IT_CC1)!=RESET)
  {
    TIM_ClearITPendingBit(MOTION_TIMER,TIM_IT_CC1);
    capture = TIM_GetCapture1(MOTION_TIMER);
    TIM_SetCompare1(MOTION_TIMER, capture + manager_motion_period);
    // call the action process
    action_process();
    // call the joint motion process
    // joint_motion_process();
    // call the walking process
    // walking_process();
    // get the target angles from all modules
    manager_get_target_position();
    // balance the robot
//    manager_balance();
    // send the motion commands to the servos
    manager_send_motion_command();
    // get the disabled servos position
//    manager_get_current_position();
  }
}

// public functions
void manager_init(uint16_t period_us)
{
  TIM_TimeBaseInitTypeDef  TIM_TimeBaseStructure;
  TIM_OCInitTypeDef  TIM_OCInitStructure;
  NVIC_InitTypeDef NVIC_InitStructure;
  uint8_t servo_ids[MANAGER_MAX_NUM_SERVOS];
  uint8_t i,num=0;
  uint16_t model;

  RCC_APB1PeriphClockCmd(MOTION_TIMER_CLK,ENABLE);

  // enable power to the servos
  dyn_master_enable_power();
  delay_ms(300);

  // detect the servos connected 
  dyn_master_scan(&num,servo_ids); 
  ram_data[BIOLOID_MM_NUM_SERVOS]=num;
  manager_num_servos=0;
  for(i=0;i<MANAGER_MAX_NUM_SERVOS;i++)
  {
    if(i==servo_ids[manager_num_servos])
    {
      // read the model of the i-th device
      dyn_master_read_word(servo_ids[manager_num_servos],P_MODEL_NUMBER_L,&model);
      switch(model)
      {
        case SERVO_AX12A: manager_servos[i].encoder_resolution=1023;
                          manager_servos[i].gear_ratio=254;
                          manager_servos[i].max_angle=300<<7;
                          manager_servos[i].center_angle=150<<7;
                          manager_servos[i].max_speed=354;
                          break;
        case SERVO_AX12W: manager_servos[i].encoder_resolution=1023;
                          manager_servos[i].gear_ratio=32;
                          manager_servos[i].max_angle=300<<7;
                          manager_servos[i].center_angle=150<<7;
                          manager_servos[i].max_speed=2830;
                          break;
        case SERVO_AX18A: manager_servos[i].encoder_resolution=1023;
                          manager_servos[i].gear_ratio=254;
                          manager_servos[i].max_angle=300<<7;
                          manager_servos[i].center_angle=150<<7;
                          manager_servos[i].max_speed=582;
                          break;
        case SERVO_MX28: manager_servos[i].encoder_resolution=4095;
                         manager_servos[i].gear_ratio=193;
                         manager_servos[i].max_angle=360<<7;
                         manager_servos[i].center_angle=180<<7;
                         manager_servos[i].max_speed=330;
                         break;
        case SERVO_RX24F: manager_servos[i].encoder_resolution=1023;
                          manager_servos[i].gear_ratio=193;
                          manager_servos[i].max_angle=300<<7;
                          manager_servos[i].center_angle=150<<7;
                          manager_servos[i].max_speed=756;
                          break;
        case SERVO_RX28: manager_servos[i].encoder_resolution=1023;
                         manager_servos[i].gear_ratio=193;
                         manager_servos[i].max_angle=300<<7;
                         manager_servos[i].center_angle=150<<7;
                         manager_servos[i].max_speed=402;
                         break;
        case SERVO_RX64: manager_servos[i].encoder_resolution=1023;
                         manager_servos[i].gear_ratio=200;
                         manager_servos[i].max_angle=300<<7;
                         manager_servos[i].center_angle=150<<7;
                         manager_servos[i].max_speed=294;
                         break;
        case SERVO_MX64: manager_servos[i].encoder_resolution=4095;
                         manager_servos[i].gear_ratio=200;
                         manager_servos[i].max_angle=360<<7;
                         manager_servos[i].center_angle=180<<7;
                         manager_servos[i].max_speed=378;
                         break;
        case SERVO_EX106: manager_servos[i].encoder_resolution=1023;
                          manager_servos[i].gear_ratio=184;
                          manager_servos[i].max_angle=250<<7;
                          manager_servos[i].center_angle=125<<7;
                          manager_servos[i].max_speed=414;
                          break;
        case SERVO_MX106: manager_servos[i].encoder_resolution=4095;
                          manager_servos[i].gear_ratio=225;
                          manager_servos[i].max_angle=360<<7;
                          manager_servos[i].center_angle=180<<7;
                          manager_servos[i].max_speed=270;
                          break;
        default: break; 
      }
      manager_servos[i].id=servo_ids[manager_num_servos];
      manager_servos[i].model=model;
      manager_servos[i].module=MM_NONE;
      manager_servos[i].enabled=0x00;
      // get the servo's current position
      dyn_master_read_word(manager_servos[i].id,P_PRESENT_POSITION_L,&manager_servos[i].current_value);
      manager_servos[i].current_angle=manager_value_to_angle(i,manager_servos[i].current_value);
      ram_data[BIOLOID_MM_SERVO0_CUR_POS_L+i*2]=(manager_servos[i].current_angle&0xFF);
      ram_data[BIOLOID_MM_SERVO0_CUR_POS_H+i*2]=(manager_servos[i].current_angle>>8);
      // set the action current angles
      action_current_angles[i]=manager_servos[i].current_angle<<9;
      manager_num_servos++;
    }
    else
    {
      manager_servos[i].id=i;
      manager_servos[i].model=0x0000;
      manager_servos[i].module=MM_NONE;
      manager_servos[i].encoder_resolution=0;
      manager_servos[i].gear_ratio=0;
      manager_servos[i].max_angle=0;
      manager_servos[i].center_angle=0;
      manager_servos[i].max_speed=0;
      manager_servos[i].current_value=0;
      manager_servos[i].current_angle=0;
      manager_servos[i].enabled=0x00;
    }
  }
  dyn_master_disable_power();

  // initialize the timer interrupts
  NVIC_InitStructure.NVIC_IRQChannel = MOTION_TIMER_IRQn;
  NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 2;
  NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
  NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
  NVIC_Init(&NVIC_InitStructure);

  /* motion timer configuration */
  TIM_TimeBaseStructure.TIM_Period = 0xFFFF;
  TIM_TimeBaseStructure.TIM_Prescaler = 72;
  TIM_TimeBaseStructure.TIM_ClockDivision = 0;
  TIM_TimeBaseStructure.TIM_CounterMode = TIM_CounterMode_Up;
  TIM_TimeBaseInit(MOTION_TIMER,&TIM_TimeBaseStructure);
  TIM_SetClockDivision(MOTION_TIMER,TIM_CKD_DIV1);

  TIM_OCInitStructure.TIM_OCMode = TIM_OCMode_Timing;
  TIM_OCInitStructure.TIM_OutputState = TIM_OutputState_Disable;
  TIM_OCInitStructure.TIM_OCPolarity = TIM_OCPolarity_High;
  manager_motion_period=period_us;
  TIM_OCInitStructure.TIM_Pulse = manager_motion_period;
  TIM_OC1Init(MOTION_TIMER, &TIM_OCInitStructure);
  TIM_OC1PreloadConfig(MOTION_TIMER, TIM_OCPreload_Disable);

  ram_clear_bit(BIOLOID_MM_STATUS,0);
}

uint16_t manager_get_period(void)
{
  return manager_motion_period;
}

void manager_set_period(uint16_t period_us)
{
  manager_motion_period=period_us;
}

inline void manager_enable(void)
{
  TIM_Cmd(MOTION_TIMER, ENABLE);
  TIM_ITConfig(MOTION_TIMER, TIM_IT_CC1, ENABLE);
  ram_set_bit(BIOLOID_MM_STATUS,0);
}

inline void manager_disable(void)
{
  TIM_Cmd(MOTION_TIMER, DISABLE);
  TIM_ITConfig(MOTION_TIMER, TIM_IT_CC1, DISABLE);
  ram_clear_bit(BIOLOID_MM_STATUS,0);
}

inline uint8_t manager_is_enabled(void)
{
  return ram_data[BIOLOID_MM_STATUS]&0x01;
}

inline uint8_t manager_get_num_servos(void)
{
  return manager_num_servos;
}

void manager_set_module(uint8_t servo_id,TModules module)
{
  if(servo_id>=0 && servo_id<MANAGER_MAX_NUM_SERVOS)
    manager_servos[servo_id].module=module;
}

inline TModules manager_get_module(uint8_t servo_id)
{
  if(servo_id>=0 && servo_id<MANAGER_MAX_NUM_SERVOS)
    return manager_servos[servo_id].module;
  else
    return MM_NONE;
}

inline void manager_enable_servo(uint8_t servo_id)
{
  if(servo_id>=0 && servo_id<MANAGER_MAX_NUM_SERVOS)
    manager_servos[servo_id].enabled=0x01;
}

inline void manager_disable_servo(uint8_t servo_id)
{
  if(servo_id>=0 && servo_id<MANAGER_MAX_NUM_SERVOS)
    manager_servos[servo_id].enabled=0x00;
}

void manager_disable_servos(void)
{
  uint8_t servo_ids[MANAGER_MAX_NUM_SERVOS];
  uint8_t data[256];
  uint8_t i,num=0;

  for(i=0;i<MANAGER_MAX_NUM_SERVOS;i++)
  {
    if(!manager_servos[i].enabled)
    {
      servo_ids[num]=manager_servos[i].id;
      data[num]=0x00;
      num++;
    }
  }
  if(num>0)
    dyn_master_sync_write(num,servo_ids,P_TORQUE_ENABLE,1,data);
}

inline uint8_t manager_is_servo_enabled(uint8_t servo_id)
{
  if(servo_id>=0 && servo_id<MANAGER_MAX_NUM_SERVOS)
    return manager_servos[servo_id].enabled;
  else
    return 0x00;
}

